/***
*		stretch.c -- make clip fit into a time tool
*		 1993, 1994, 1997 Another Option
***/

/*
============================================================================
|    DATE    | WHAT
============================================================================
   19 Jul 93 | created
	20 Jul 93 | fixed and added more list handling when doing a clip
	21 Jul 93 | removed reading edit markers to get target time; preserved
	            in stretchmarks.c, but can't figure out how to know when to
					read tracks window's or a track's markers; added MIDIONLY
	15 Nov 93 | new window title
	23 Jan 94 | new icon and name; must have processevent()!
	13 Feb 94 | slightly different ui; added smpte time; changed to Rope gadgets
	13 Sep 97 | sync gadget image for time toggle gadget; limit duration scaling
	          | to only notes; limit minimum durations to 1
   16 Sep 97 | added calculating duration of events for scaling
	22 Oct 97 | text[] 1 char too short in edittool()!? something changed; now
	          | smpte conversion is way fucked
*/

#define DBUG 0
//#define FILEDUMP 1

#include <exec/exec.h>
#include <libraries/dos.h>
#include <string.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <stdio.h>

#include "bars.h"
#include "edit.h"
#include "bplib.h"
#include "barsmacros.h"
#include "proto/stretch_protos.h"
#include "timewarpw.h"
#include "/icons/TimeWarp.h"
#include "timetogglegadget.h"
#include "toolid.h"
#include "debug.h"

static UBYTE const versiontag[] = "$VER: TimeWarp 40.21 (22.10.97)";

struct StretchTool {
	struct Tool tool;
	long target_time; /* time to fill */
	ULONG flags;
};

/* flags */
#define ADJUST_DUR 1
#define MIDI_ONLY  2
#define SMPTE_TIME 4
#define CALC_TIME  8 // don't rely on clip->begin & clip->end; add it up

#define TOOL_SIZE sizeof(struct StretchTool) /* for ToolMaster */

extern struct Functions * far functions;

/* use this multiplier to avoid floating point when scaling */
#define RESOLUTION 10000

/* maxlong/resolution = biggest time span -- this is 559.0.184 */
#define MAX_TIMESPAN (4294967295 / RESOLUTION)

/*==========================================================================
* inittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: sets tool defaults
*=========================================================================*/

static void __regargs inittool(struct StretchTool *tool)

{
	tool->target_time = 3072; /* 4 bars */
	tool->flags = ADJUST_DUR | CALC_TIME;
	tool->tool.touched |= TOUCH_INIT;
}

/*==========================================================================
* scalelist
*
*  INPUT: pointer to list of Events, time of first event, scalar, 
*         flag to indicate whether to scale durations
* RETURN: none
*
* EFFECT: multiplies times and maybe durations by scalar
*=========================================================================*/

static void __regargs scalelist(struct Event *list, long start, ULONG scalar,
                                UBYTE dodur)

{
	struct NoteEvent *event = (struct NoteEvent *)list->next; /* leave the first time alone */

	if (dodur && IsNoteOn(list)) {
		((struct NoteEvent *)list)->duration = 
                       (((struct NoteEvent *)list)->duration * scalar)/RESOLUTION;
		if (((struct NoteEvent *)list)->duration < 1) {
			((struct NoteEvent *)list)->duration = 1;
		}
	}
	for ( ; event ; event=event->next ) {
//		event->time = ((event->time - list->time) * scalar)/RESOLUTION + list->time;
		event->time = ((event->time - start) * scalar)/RESOLUTION + start;
		if (dodur && IsNoteOn(event)) {
			event->duration = (event->duration * scalar)/RESOLUTION;
			if (event->duration < 1) event->duration = 1;
		}
	}
}
		
/*==========================================================================
* stretchclip
*
*  INPUT: pointer to Clip, scalar, tool flags
* RETURN: none
*
* EFFECT: stretches event lists in the clip
*=========================================================================*/

static void __regargs stretchclip(struct Clip *clip, long start, ULONG scalar,
                                  ULONG flags)

{
	struct DynamicsEvent *event;
	struct DynamicsEvent *prev;
	long lasttime;

//	scalelist(clip->events.first,scalar,flags & ADJUST_DUR);
	scalelist(clip->events.first,clip->events.first->time,scalar,flags & ADJUST_DUR);
	if (!(flags & MIDI_ONLY)) {
		if (clip->chords.first)
			scalelist(clip->chords.first,clip->chords.first->time,scalar,0);
		if (clip->keys.first)
			scalelist(clip->keys.first,clip->keys.first->time,scalar,0);
		if (clip->lyrics.first)
			scalelist(clip->lyrics.first,clip->lyrics.first->time,scalar,0);
		if (clip->rhythm.first)
			scalelist(clip->rhythm.first,clip->rhythm.first->time,scalar,0);
		if (clip->dynamics.first) {
			scalelist(clip->dynamics.first,clip->dynamics.first->time,scalar,0);
			/* redo the times in these events */
			lasttime = clip->dynamics.first->time;
			prev = (struct DynamicsEvent *)clip->dynamics.first;
			event = (struct DynamicsEvent *)clip->dynamics.first->next;
			for ( ; event ; event=event->next ) {
				event->lasttime = lasttime;
				prev->nexttime = event->time;
				lasttime = event->time;
				prev = event;
			}
		}
/*! TIMESIGLIST STILL TO DO.  but it looks hard */
	}
}

/*==========================================================================
* processlist
*=========================================================================*/

static struct Event * __stdargs __saveds processlist(struct StretchTool *tool,
                                                     struct NoteEvent *list,
                                                     long start, long stop)

{
	struct NoteEvent *end = NULL;
	ULONG scalar;

	if (!(tool->tool.touched)) inittool(tool);
	if (list) {
		scalar = (ULONG)(tool->target_time * RESOLUTION);
		if (tool->flags & CALC_TIME) { /* use actual duration of list events */
			for (end=list ; end->next && (end->time < stop) ; end=end->next)
				;
			if (end != list) {
				scalar /= IsNoteOn(end) ? 
							 end->time + end->duration - list->time :
							 end->time - list->time;
			}
		}
		else { /* use the edit flags */
			if (start < 0) start = 0;
			scalar /= stop - start;
		}
		if (end != list) {
			scalelist((struct Event *)list, 
			          (tool->flags & CALC_TIME) ? list->time : start, 
						 scalar, tool->flags & ADJUST_DUR);
		}
	}
	return((struct Event *)list);
}

/*==========================================================================
* stretchclipboard
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: gets a clip from the clipboard; passes the eventlist to scalelist;
*         posts the stretched clip.
*=========================================================================*/


/*! if a clip from the track window is really a ClipList, then this needs to
    get a pointer to a Clip from edittool(), which will need to get the clip
	 or list from the clipboard, and feed them to stretchclipboard() 
!*/

static void __regargs stretchclipboard(struct StretchTool *tool)
{
#pragma msg 104 ignore push
	struct Clip volatile c;
	struct Clip *clip = &c;
#pragma msg 104 ignore pop
	struct NoteEvent *end;
	ULONG scalar;
	long firsttime;
	long lasttime;

	if (functions->clipboard(CLIP_GET,clip) == 1) {
		firsttime = 0x7FFFFFFF;
		if (clip->events.first) firsttime = clip->events.first->time;
		if (!tool->flags & MIDI_ONLY) {
			if (clip->chords.first) {
				if (clip->chords.first->time < firsttime) firsttime = clip->chords.first->time;
			}
			if (clip->keys.first) {
				if (clip->keys.first->time < firsttime) firsttime = clip->keys.first->time;
			}
			if (clip->lyrics.first) {
				if (clip->lyrics.first->time < firsttime) firsttime = clip->lyrics.first->time;
			}
			if (clip->rhythm.first) {
				if (clip->rhythm.first->time < firsttime) firsttime = clip->rhythm.first->time;
			}
			if (clip->dynamics.first) {
				if (clip->dynamics.first->time < firsttime) firsttime = clip->dynamics.first->time;
			}
		}
//dprintf("firsttime=%ld\n",firsttime);
		if (tool->flags & CALC_TIME) {
			lasttime = -1;
			end = (struct NoteEvent *)listend((struct MinNode *)clip->events.first);
/*			end = (struct NoteEvent *)(clip->events.first);
			for( ; end ; end = end->next ) {
				if (end->time > lasttime) lasttime = end->time;
				if (IsNoteOn(end)) lasttime += end->duration;
			}
*/
			if (end) {
				if (end->time > lasttime) lasttime = end->time;
				if (IsNoteOn(end)) lasttime += end->duration;
			}
			if (!tool->flags & MIDI_ONLY) { // find the time of the really last event
				end = (struct NoteEvent *)listend((struct MinNode *)clip->chords.first);
				if (end) {
					if (end->time > lasttime) lasttime = end->time;
				}
				end = (struct NoteEvent *)listend((struct MinNode *)clip->keys.first);
				if (end) {
					if (end->time > lasttime) lasttime = end->time;
				}
				end = (struct NoteEvent *)listend((struct MinNode *)clip->lyrics.first);
				if (end) {
					if (end->time > lasttime) lasttime = end->time;
				}
				end = (struct NoteEvent *)listend((struct MinNode *)clip->rhythm.first);
				if (end) {
					if (end->time > lasttime) lasttime = end->time;
				}
				end = (struct NoteEvent *)listend((struct MinNode *)clip->dynamics.first);
				if (end) {
					if (end->time > lasttime) lasttime = end->time;
				}
			}
			scalar = lasttime - firsttime;
//dprintf("lasttime=%ld\n",lasttime);
		}
		else {
//dprintf("clip->end=%ld clip->begin=%ld\n",clip->end,clip->begin);
			scalar = clip->end - clip->begin;
		}
//dprintf("scalar(dur)=%ld\n",scalar);
		if ((scalar > 0) && (firsttime < 0x7FFFFFFF)) {
			scalar = (ULONG)(tool->target_time * RESOLUTION) / scalar;
//dprintf("scalar=%ld\n",scalar);
			if (scalar) {
				stretchclip(clip,firsttime,scalar,tool->flags);
				/*	damned clipboard assigns its own name, but demands a clip.name */
				functions->freestring(clip->name);
#pragma msg 104 ignore push
				clip->name = functions->allocstring("B"); 
#pragma msg 104 ignore pop
				clip->end = clip->begin + tool->target_time;
				functions->clipboard(CLIP_PUT,clip); /* post the stretched clip */
			}
		}
	}
}


#define tw tool->tool.window

/*==========================================================================
* edittool
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: ui handler
*=========================================================================*/

static void __stdargs __saveds edittool(struct StretchTool *tool)

{
	struct IntuiMessage *message;
	long class; 
	long code;
	short mousex, mousey;
	struct Gadget *gadget;
	struct NewWindow *nw;
	long rope_return;
	char text[13];
	UBYTE refresh = 3;

	if (!(tool->tool.touched & TOUCH_INIT)) inittool(tool);

#pragma msg 104 ignore push
	if (!(nw=OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,20)))
		return;
#pragma msg 104 ignore pop
	AddTimeToggleGadget(tw);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	EmbossOnGList(tw,&GadgetList1,GCOUNT,1);
	SetMenuStrip(tw,&Menu1);
	if (gadget = functions->GetGadget(tw,TIME)) {
		functions->InitRopeGadget(tw,gadget,TIME,TIME);
		((struct RopeGadget *)(gadget->UserData))->flags |= ROPE_LENGTH;
	}
	// sync imagery for time toggle gadget
	functions->SelectGadget(tw,TIMETOGGLE_GADGET_ID,tool->flags & SMPTE_TIME);
	functions->RefreshGadget(tw,TIMETOGGLE_GADGET_ID);

	for ( ; ; ) {
		if (refresh & 1) {
			functions->SelectEmbossed(tw,ADJUST,tool->flags & ADJUST_DUR);
			functions->SelectEmbossed(tw,MIDI,tool->flags & MIDI_ONLY);
			functions->SelectEmbossed(tw,CALCULATE,tool->flags & CALC_TIME);
			DrawEmbossedGList(tw,&GadgetList1,GCOUNT);
			refresh ^= 1;
		}
		if (refresh & 2) {
//dprintf("tool->target_time=%ld  ",tool->target_time);
			if (tool->flags & SMPTE_TIME) {
				functions->frametostring(functions->timetoframe(tool->target_time),text);
//dprintf("smpte: %s\n",text);
			}
			else {
				functions->lengthtostring(tool->tool.clip,0,tool->target_time,text);
//dprintf("music: %s\n",text);
			}
			functions->SetRopeInfo(tw,TIME,text);
			RefreshGList(functions->GetGadget(tw,TIME),tw,NULL,1);
			refresh=0;
		}
		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		class = message->Class;
		code = message->Code;
		mousex = message->MouseX;
		mousey = message->MouseY;
		gadget = (struct Gadget *)message->IAddress;
		ReplyMsg((struct Message *)message);
		class = functions->SystemGadgets(tw,class,gadget,code);

		if (class == CLOSEWINDOW)
			break;
		else if (class == GADGETDOWN) {
			if (gadget->GadgetID == TIME) {
				rope_return = functions->HandleRopeGadget(tw,gadget,mousex,mousey);
//dprintf("rope_return=%ld\n",rope_return);
				if (rope_return != -1) {
					tool->target_time = (rope_return <= MAX_TIMESPAN) ?
											  rope_return : MAX_TIMESPAN;
					if (tool->target_time < 1) tool->target_time = 1;
					refresh=2;
				}
			}
			else if (gadget->GadgetID == TIMETOGGLE_GADGET_ID) {
				tool->flags ^= SMPTE_TIME;
				refresh=2;
			}
		}
		else if (class == GADGETUP) {
			switch (gadget->GadgetID) {
				case GRAB :
					stretchclipboard(tool);
					break;
				case CALCULATE :
					tool->flags ^= CALC_TIME;
					refresh=1;
					break;
				case ADJUST :
					tool->flags ^= ADJUST_DUR;
					refresh=1;
					break;
				case MIDI :
					tool->flags ^= MIDI_ONLY;
					refresh = 1;
					break;
			}
		}
	}

	EmbossOffGList(tw,&GadgetList1,GCOUNT);
	functions->FreeRopeGadget(functions->GetGadget(tw,TIME));
	NewCloseEmbossedWindow(&tool->tool,nw);
}

/********************************************/
static struct Event * __stdargs __saveds processevent(struct Event *event)
{
	event->tool = event->tool->next;
	return(event);
}


static struct ToolMaster master;

/*==========================================================================
* inittoolmaster
*=========================================================================*/

struct ToolMaster * __stdargs __saveds inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = TOOL_SIZE;
	master.toolid = STRC_TOOLID;
	master.edittool = edittool;
	master.processevent = processevent;
	master.processlist = processlist;
	master.image = &TimeWarpIcon;
	strcpy(master.name,WTITLE);
	master.tooltype = TOOL_NORMAL;

	return(&master);
}
